/*
 * File: dw_page.c
 *
 * Copyright (C) 1997 Raph Levien <raph@acm.org>
 * Copyright (C) 1999 Luca Rota <drake@freemail.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

/*
 * This module contains the Dw_page widget, which is the "back end" to
 * Web text widgets including html. 
 */

#define TRACE_GTK_METHODS
#undef NO_WINDOW
#define HUGE_WINDOW
#define DW_PAGE_MAX_WINDOW_SIZE 32760

#include <ctype.h>
#include <string.h>
#include <stdio.h>

#include <gtk/gtk.h>

#include "list.h"
#include "dw_page.h"
#include "dw_marshal.h"
#include "dw_gtk_viewport.h"

#include "prefs.h"

//#define DEBUG_LEVEL 10
#include "debug.h"

static void Dw_page_init          (DwPage *page);
static void Dw_page_class_init    (DwPageClass *klass);

static void Dw_page_destroy       (GtkObject *object);

static void Dw_page_size_request      (DwWidget *widget,
                                      DwRequisition *requisition);
static void Dw_page_fast_size_request (DwWidget *widget,
                                       DwRequisition *requisition);
static void Dw_page_get_extremes      (DwWidget *widget,
                                       DwExtremes *extremes);
static void Dw_page_size_allocate     (DwWidget *widget,
                                       DwAllocation *allocation);
static void Dw_page_set_width         (DwWidget *widget,
                                       gint32 width);
static void Dw_page_set_ascent        (DwWidget *widget,
                                       gint32 ascent);
static void Dw_page_set_descent       (DwWidget *widget,
                                       gint32 descent);
static void Dw_page_draw              (DwWidget *page,
                                       DwRectangle *area,
                                       GdkEventExpose *event);
static gint Dw_page_button_press      (DwWidget *widget,
                                       gint32 x,
                                       gint32 y,
                                       GdkEventButton *event);
static gint Dw_page_button_release    (DwWidget *widget,
                                       gint32 x,
                                       gint32 y,
                                       GdkEventButton *event);
static gint Dw_page_motion_notify     (DwWidget *widget,
                                       gint32 x,
                                       gint32 y,
                                       GdkEventMotion *event);
static gint Dw_page_leave_notify      (DwWidget *widget,
                                       GdkEventMotion *event);


static void Dw_page_add               (DwContainer *container,
                                       DwWidget *widget);
static void Dw_page_remove            (DwContainer *container,
                                       DwWidget *widget);
static void Dw_page_forall            (DwContainer *container,
                                       DwCallback callback,
                                       gpointer callback_data);
static gint Dw_page_findtext          (DwContainer *container,
                                       gpointer FP, gpointer KP,
                                       gchar *NewKey);

static void Dw_page_rewrap            (DwPage *page);

/*
 * Returns the x offset (the indentation plus any offset needed for
 * centering or right justification) for the line. The offset returned
 * is relative to the page *content* (i.e. without border etc.).
 */
#define Dw_page_line_x_offset(page, line) \
   ( (line)->first ? \
        (page)->words[(line)->first_word].style->left_indent_first : \
        (page)->words[(line)->first_word].style->left_indent_rest )

/*
 * Like Dw_style_box_offset_x, but relative to the allocation (i.e.
 * including border etc.).
 */
#define Dw_page_line_total_x_offset(page, line) \
   ( Dw_page_line_x_offset (page, line) + \
     Dw_style_box_offset_x (((DwWidget*)(page))->style) )


enum
{
   LINK_ENTERED,
   LINK_PRESSED,
   LINK_RELEASED,
   LINK_CLICKED,
   LAST_SIGNAL
};

static DwContainerClass *parent_class;
static guint page_signals[LAST_SIGNAL] = { 0 };


/*
 * Return the type of DwPage
 */
GtkType a_Dw_page_get_type (void)
{
   static GtkType type = 0;

   if (!type) {
      GtkTypeInfo info = {
         "DwPage",
         sizeof (DwPage),
         sizeof (DwPageClass),
         (GtkClassInitFunc) Dw_page_class_init,
         (GtkObjectInitFunc) Dw_page_init,
         (GtkArgSetFunc) NULL,
         (GtkArgGetFunc) NULL,
         (GtkClassInitFunc) NULL
      };
      
      type = gtk_type_unique (DW_TYPE_CONTAINER, &info);
   }
   
   return type;
}


/*
 * Create a new DwPage
 */
DwWidget* a_Dw_page_new (void)
{
   GtkObject *object;

   object = gtk_object_new (DW_TYPE_PAGE, NULL);
   return DW_WIDGET (object);
}


/*
 * Initialize a DwPage
 */
static void Dw_page_init (DwPage *page)
{
   DW_WIDGET_SET_FLAGS (page, DW_USES_HINTS);
   
   /* 
    * The initial sizes of page->lines and page->words should not be
    * too high, since this will waste much memory with tables
    * containing many small cells. The few more calls to realloc
    * should not decrease the speed considerably.
    * (Current setting is for minimal memory usage. An interesting fact
    * is that high values decrease speed due to memory handling overhead!)
    * todo: Some tests would be useful.
    */
   page->num_lines_max = 1; /* 2 */
   page->num_lines = 0;
   page->lines = NULL;      /* g_new(DwPageLine, page->num_lines_max); */

   page->num_words_max = 1; /* 8 */
   page->num_words = 0;
   page->current_word = 0;
   page->words = NULL;      /* g_new(DwPageWord, page->num_words_max); */

   page->last_line_max_width = 0;

   page->hover_link = -1;

   /* random values */
   page->avail_width = 100;
   page->avail_ascent = 100;
   page->avail_descent = 0;
}

/*
 * Initialize the DwPage's class
 */
static void Dw_page_class_init (DwPageClass *klass)
{
   GtkObjectClass *object_class;
   DwWidgetClass *widget_class;
   DwContainerClass *container_class;

   object_class = (GtkObjectClass*) klass;
   widget_class = (DwWidgetClass*) klass;
   container_class = (DwContainerClass*) klass;
   parent_class = gtk_type_class (DW_TYPE_CONTAINER);

   /* todo: the link_xxx signals should return boolean values, which are
    * then returned by the event signal function of DwPage. But I don't
    * know that much about Gtk+ signals :-(
    * --SG 
    */
   page_signals[LINK_ENTERED] =
      gtk_signal_new ("link_entered",
                      GTK_RUN_FIRST,
                      object_class->type,
                      GTK_SIGNAL_OFFSET (DwPageClass, link_entered),
                      Dw_marshal_link_enter,
                      GTK_TYPE_NONE,
                      3, GTK_TYPE_INT, GTK_TYPE_INT, GTK_TYPE_INT);
   page_signals[LINK_PRESSED] =
      gtk_signal_new ("link_pressed",
                      GTK_RUN_FIRST,
                      object_class->type,
                      GTK_SIGNAL_OFFSET (DwPageClass, link_pressed),
                      Dw_marshal_link_button,
                      GTK_TYPE_NONE,
                      4, GTK_TYPE_INT, GTK_TYPE_INT, GTK_TYPE_INT,
                      GTK_TYPE_GDK_EVENT);
   page_signals[LINK_RELEASED] =
      gtk_signal_new ("link_released",
                      GTK_RUN_FIRST,
                      object_class->type,
                      GTK_SIGNAL_OFFSET (DwPageClass, link_released),
                      Dw_marshal_link_button,
                      GTK_TYPE_NONE,
                      4, GTK_TYPE_INT, GTK_TYPE_INT, GTK_TYPE_INT,
                      GTK_TYPE_GDK_EVENT);
   page_signals[LINK_CLICKED] =
      gtk_signal_new ("link_clicked",
                      GTK_RUN_FIRST,
                      object_class->type,
                      GTK_SIGNAL_OFFSET (DwPageClass, link_clicked),
                      Dw_marshal_link_button,
                      GTK_TYPE_NONE,
                      4, GTK_TYPE_INT, GTK_TYPE_INT, GTK_TYPE_INT,
                      GTK_TYPE_GDK_EVENT);
   gtk_object_class_add_signals (object_class, page_signals, LAST_SIGNAL);
   
   object_class->destroy = Dw_page_destroy;

   widget_class->size_request = Dw_page_size_request;
   widget_class->fast_size_request = Dw_page_fast_size_request;
   widget_class->get_extremes = Dw_page_get_extremes;
   widget_class->size_allocate = Dw_page_size_allocate;
   widget_class->set_width = Dw_page_set_width;
   widget_class->set_ascent = Dw_page_set_ascent;
   widget_class->set_descent = Dw_page_set_descent;
   widget_class->draw = Dw_page_draw;
   widget_class->button_press_event = Dw_page_button_press;
   widget_class->button_release_event = Dw_page_button_release;
   widget_class->motion_notify_event = Dw_page_motion_notify;
   widget_class->leave_notify_event = Dw_page_leave_notify;

   container_class->add = Dw_page_add;
   container_class->remove = Dw_page_remove;
   container_class->forall = Dw_page_forall;
   container_class->findtext = Dw_page_findtext;

   klass->link_entered = NULL;
   klass->link_pressed = NULL;
   klass->link_released = NULL;
   klass->link_clicked = NULL;
}


/*
 * Destroy the page (standard Gtk+ function)
 */
static void Dw_page_destroy (GtkObject *object)
{
   DwPage *page;
   DwPageWord *word;
   gint i;

   DEBUG_MSG(10, "Dw_page_destroy\n");
   
   page = DW_PAGE (object);

   for (i = 0; i < page->num_words; i++) {
      word = &page->words[i];
      if (word->content_type == DW_PAGE_CONTENT_WIDGET)
         gtk_object_unref(GTK_OBJECT(word->content.widget));
      else if (word->content_type == DW_PAGE_CONTENT_TEXT)
         g_free(word->content.text);
      else if (word->content_type == DW_PAGE_CONTENT_ANCHOR)
         g_free(word->content.anchor);
      
      a_Dw_style_unref(word->style);
   }

   g_free (page->lines);
   g_free (page->words);

   /* make sure we don't own widgets anymore; Is this necessary? */
   page->num_words = 0; 
   page->num_lines = 0;

   (* GTK_OBJECT_CLASS(parent_class)->destroy) (object);
}


/*
 * Standard Dw function
 */
static void Dw_page_size_request (DwWidget *widget,
                                  DwRequisition *requisition)
{
   Dw_page_rewrap(DW_PAGE (widget));
   Dw_page_fast_size_request (widget, requisition); /* does the rest */
}


/*
 * Standard Dw function
 */
static void Dw_page_fast_size_request (DwWidget *widget,
                                       DwRequisition *requisition)
{
   /*
    * This (faster!) function is only called, when all there is no
    * rewrap necessary.
    */
   DwPage *page;

   page = DW_PAGE (widget);

   requisition->width = page->real_width +
      Dw_style_box_diff_width (widget->style);

   if (page->num_lines > 0)
      requisition->ascent =
         page->lines[page->num_lines - 1].y_top +
         page->lines[page->num_lines - 1].size.ascent +
         page->lines[page->num_lines - 1].size.descent -
         Dw_style_box_offset_y (widget->style);
   else
      requisition->ascent = 0;

   requisition->ascent += Dw_style_box_diff_height (widget->style);

   /* todo: requisition->descent could get a useful value (as in TeX),
      but I don't have any idea where this will really be needed ;-) */
   requisition->descent = 0;
}


/*
 * Standard Dw function
 */
static void Dw_page_get_extremes (DwWidget *widget,
                                  DwExtremes *extremes)
{
   /* todo: This could perhaps be optimized,
      similar to Dw_page_size_request. */
   DwPage *page;
   DwExtremes word_extremes;
   DwPageLine *line;
   DwPageWord *word;
   gint line_index, word_index;
   gint32 para_min_width, indent_para_min_width;
   gint32 para_max_width, indent_para_max_width;
   gboolean infinitive_max_width, nowrap;

   page = DW_PAGE (widget);
   extremes->min_width = 0;
   extremes->max_width = 0;
   para_min_width = 0;
   para_max_width = 0;
   infinitive_max_width = FALSE;
   
   DEBUG_MSG (4, "Dw_page_get_extremes\n");

   for (line_index = 0; line_index < page->num_lines; line_index++) {
      line = &page->lines[line_index];
      nowrap = page->words[line->first_word].style->nowrap;

      for (word_index = line->first_word; word_index < line->last_word;
           word_index++) {
         word = &page->words[word_index];
      
         if (word->content_type == DW_PAGE_CONTENT_WIDGET) {
            if (DW_WIDGET_USES_HINTS (word->content.widget))
               a_Dw_widget_get_extremes (word->content.widget, &word_extremes);
            else {
               if (DW_STYLE_IS_PERCENTAGE
                   (word->content.widget->style->width)) {
                  word_extremes.min_width = 0;
                  word_extremes.max_width = 0;

                  if (DW_WIDGET_HAS_CONTENT (word->content.widget))
                     infinitive_max_width = TRUE;
               } else if (DW_STYLE_IS_LENGTH
                          (word->content.widget->style->width))
                  word_extremes.min_width = word_extremes.max_width =
                     DW_STYLE_GET_LENGTH (word->content.widget->style->width,
                                          word->style->font);
               else
                  a_Dw_widget_get_extremes (word->content.widget,
                                            &word_extremes);
            }
         } else {
            word_extremes.min_width = word->size.width;
            word_extremes.max_width = word->size.width;
         }
         
         if (nowrap)
            para_min_width += word_extremes.min_width + word->x_space;
         else {
            /* Assumed that there will only be left_indent... */
            if (extremes->min_width <
                word_extremes.min_width + word->style->left_indent_first)
               extremes->min_width =
                  word_extremes.min_width + word->style->left_indent_first;
         }

         para_max_width += word_extremes.max_width + word->x_space;
         
         DEBUG_MSG (1, "      word \"%s\": max_width = %d\n",
                    word->content_type == DW_PAGE_CONTENT_TEXT ?
                    word->content.text : "<no text>",
                    word_extremes.max_width);
      }

      if (line->last_word > line->first_word &&
          (line->hard || line_index == page->num_lines - 1)) {
         word = &page->words[line->last_word - 1];
         indent_para_max_width =
            para_max_width + word->style->left_indent_first - word->x_space;
         
         DEBUG_MSG (2, "   indent_para_max_width = %d, "
                    "after word \"%s\" (indent = %d)\n",
                    indent_para_max_width,
                    word->content_type == DW_PAGE_CONTENT_TEXT ?
                    word->content.text : "<no text>",
                    word->style->left_indent_first);
         
         if (extremes->max_width < indent_para_max_width)
            extremes->max_width = indent_para_max_width;

         if (nowrap) {
            indent_para_min_width =
               para_min_width + word->style->left_indent_first - word->x_space;
            if (extremes->min_width < indent_para_min_width)
               extremes->min_width = indent_para_min_width;
         }
         
         para_min_width = 0;
         para_max_width = 0;
      }
   }

   if (infinitive_max_width)
      /* Return a random large value. */
      extremes->max_width = DW_INFINITY;

   DEBUG_MSG (3, "   Result: %d, %d\n",
              extremes->min_width, extremes->max_width);

   extremes->min_width += Dw_style_box_diff_width (widget->style);
   extremes->max_width += Dw_style_box_diff_width (widget->style);
}


/*
 * Standard Dw function
 */
static void Dw_page_size_allocate (DwWidget *widget,
                                   DwAllocation *allocation)
{
   DwPage *page;
   int line_index, word_index;
   DwPageLine *line;
   DwPageWord *word;
   int x_cursor;
   DwAllocation child_allocation;

   page = DW_PAGE (widget);

   //g_print ("Dw_page_size_allocate\n");

   for (line_index = 0; line_index < page->num_lines; line_index++) {
      line = &(page->lines[line_index]);
      x_cursor = Dw_page_line_total_x_offset(page, line);

      for (word_index = line->first_word; word_index < line->last_word; 
           word_index++) {
         word = &(page->words[word_index]);
         
         if (word->content_type == DW_PAGE_CONTENT_WIDGET) {
            /*g_print ("   %20s: (%d, %d)\n",
              gtk_type_name (GTK_OBJECT_TYPE
              (word->content.widget.widget)),
              x_cursor, line->y_top);*/

            /* todo: justification within the line is done here! */
            child_allocation.x = x_cursor + allocation->x;
            /* align=top:
               child_allocation.y = line->y_top + allocation->y;
            */
            /* align=bottom (base line) */
            child_allocation.y = line->y_top + allocation->y
               + (line->size.ascent - word->size.ascent);         
            child_allocation.width = word->size.width;
            child_allocation.ascent = word->size.ascent;
            child_allocation.descent = word->size.descent;
            a_Dw_widget_size_allocate (word->content.widget,
                                       &child_allocation);
         }          
         
         x_cursor += (word->size.width + word->x_space);
      }
   }   
}


/*
 * Standard Dw function
 */
static void Dw_page_set_width (DwWidget *widget,
                               gint32 width)
{
   DwPage *page;

   page = DW_PAGE (widget);

   //g_print (">>> Dw_page_set_width: %d -> %d\n", page->avail_width, width);

   /* If limit_text_width is set to YES, a queue_resize may also be
      necessary. */
   if (page->avail_width != width || prefs.limit_text_width) {
      page->avail_width = width;
      Dw_widget_queue_resize (widget);
   }
}


/*
 * Standard Dw function
 */
static void Dw_page_set_ascent (DwWidget *widget,
                                gint32 ascent)
{
   DwPage *page;

   page = DW_PAGE (widget);

   //g_print (">>> Dw_page_set_ascent: %d -> %d\n", page->avail_ascent,
   //         ascent);

   if (page->avail_ascent != ascent) {
      page->avail_ascent = ascent;
      Dw_widget_queue_resize (widget);
   }
}


/*
 * Standard Dw function
 */
static void Dw_page_set_descent (DwWidget *widget,
                                 gint32 descent)
{
   DwPage *page;

   page = DW_PAGE (widget);

   //g_print (">>> Dw_page_set_descent: %d -> %d\n", page->avail_descent,
   //         descent);

   if (page->avail_descent != descent) {
      page->avail_descent = descent;
      Dw_widget_queue_resize (widget);
   }
}


/*
 * Standard Dw function
 */
static void Dw_page_add (DwContainer *container,
                         DwWidget *widget)
{
   /* todo */
}


/*
 * Standard Dw function
 */
static void Dw_page_remove (DwContainer *container,
                            DwWidget *widget)
{
   /* todo */
}


/*
 * Standard Dw function
 */
static void Dw_page_forall (DwContainer *container,
                            DwCallback callback,
                            gpointer callback_data)
{
   DwPage *page;
   int word_index;
   DwPageWord *word;

   page = DW_PAGE (container);

   for(word_index = 0; word_index < page->num_words; word_index++) {
      word = &page->words[word_index];
         
      if (word->content_type == DW_PAGE_CONTENT_WIDGET)
         (*callback) (word->content.widget, callback_data);
   }
}


/* todo:
 * 
 * + do request_repaint for added stuff. But what level of
 * granularity? reinstate the update_begin/update_end cycle? Do at
 * line granularity when making a new line (but risk paint
 * inconsistency)? Do at word granularity but take on region overhead
 * on every word (optimize away if outside visible rectangle)?
 * 
 * + better formatting, including justification & floats
 * 
 */


/*
 * ?
 */
static void Dw_page_add_line(DwPage *page)
{
   gint num_lines;
   DwPageLine *line;

   num_lines = page->num_lines;
   a_List_add(page->lines,num_lines,sizeof(*page->lines),page->num_lines_max);
   line = &page->lines[num_lines];
   line->first_word = page->current_word;
   line->last_word = page->current_word;
   if (num_lines == 0) {
      line->y_top = Dw_style_box_offset_y (DW_WIDGET(page)->style);
   } else {
      line->y_top =
          page->lines[num_lines - 1].y_top + 
          page->lines[num_lines - 1].size.ascent +
          page->lines[num_lines - 1].size.descent +
          page->lines[num_lines - 1].y_space;
   }
   line->size.width = 0;
   line->size.ascent = 0;
   line->size.descent = 0;
   line->y_space = 0;
   line->hard = FALSE;
   line->first = (num_lines == 0 || (page->lines[num_lines - 1].hard &&
                                     page->lines[num_lines - 1].y_space > 0));
   page->num_lines++;
}

/*
 * Allocate a new word in a page structure. This routine is where word
 * wrapping happens. The newly allocated word has the size.width, x_space,
 * size.ascent, and size.descent fields filled in, but no others.
 */
static DwPageWord *Dw_page_new_word(DwPage *page,
                                    gint width,
                                    gint ascent,
                                    gint descent,
                                    DwStyle *style)
{
   DwWidget *widget = DW_WIDGET (page);
   DwPageLine *line;
   DwPageWord *word;
   DwStyle *style2;
   gboolean existing;
   gboolean new_line;
   gint line_width, set_width, nw, nl;
   gint32 avail_width;

   line = NULL;
   word = NULL;
   new_line = FALSE;
   existing = FALSE;

   avail_width =
      page->avail_width - Dw_style_box_diff_width(widget->style);
   if (prefs.limit_text_width &&
       avail_width > widget->viewport->allocation.width - 10)
      avail_width = widget->viewport->allocation.width - 10;
   
   /* Is it an existing word? */
   if(width == -1)
      existing = TRUE;

   if(existing) {
      word = &page->words[page->current_word];
      style2 = word->style;
      width = word->size.width;
      ascent = word->size.ascent;
      descent = word->size.descent;
   } else {
      style2 = style;
   }

   nl = page->num_lines;
   if ( nl == 0 ) {
      new_line = TRUE;
   } else {
      line = &page->lines[nl - 1];
      if ( line->hard ) {
         new_line = TRUE;
      } else {
         /* Calculate width of new line and see if it exceeds page width. */
         nw = line->last_word;
         if ( nw > line->first_word ) {
            set_width = avail_width - style2->right_indent;

            if ( line->first )
               set_width -= style2->left_indent_first;
            else
               set_width -= style2->left_indent_rest;

            line_width =
               line->size.width + page->words[nw - 1].x_space + width;
            if ( line_width > set_width && !style2->nowrap )
               new_line = TRUE;
         }
      }

      if (new_line)
         page->real_width = MAX(page->real_width,
                                line->size.width +
                                Dw_page_line_x_offset (page, line));
   }

   if (new_line) {
      Dw_page_add_line(page);
      line = &page->lines[nl];
   }

   nw = line->last_word;

   /* Add a new word */
   if(!existing) {
      a_List_add(page->words, nw, sizeof(*page->words), page->num_words_max);
      word = &page->words[nw];
      word->size.width = width;
      word->x_space = 0;
      word->size.ascent = ascent;
      word->size.descent = descent;
      page->num_words++;

      /* todo: this couple seems to be a workaround */
      /* Just in case the page only has one word */
      if (page->num_words == 1)
         page->real_width = width;
      /* Avoid problems with pages that only have one line of text */
      if (page->num_lines == 1)
         page->real_width += width;
   }

   /* Apply the attributes to the line */
   if (nw - line->first_word)
      line->size.width += page->words[nw - 1].x_space;
   line->size.width += width;
   line->size.ascent = MAX(line->size.ascent, ascent);
   line->size.descent = MAX(line->size.descent, descent);

   /* update the width requisitions */
   if ( nw == line->first_word && line->first )
      page->last_line_max_width = style2->left_indent_first;
   else if ( nw == line->first_word )
      page->last_line_max_width += page->words[line->first_word].x_space;
   else
      page->last_line_max_width += page->words[nw - 1].x_space;
   page->last_line_max_width += width;

   page->current_word++;
   line->last_word++;

   return word;
}

/* Performance is not as good as it could be - it could only rewrap if
 * it needs to (should be a cheap test: the line is ok if it fits
 * within the page->width and either it's hard or the width plus the
 * width of the first word on the next line (plus the spacing of the
 * last word) exceed the page->width). This would save on memory
 * allocation too.
 * 
 * But hey, it works!
 */


static void Dw_page_calc_widget_size (DwPage *page,
                                      DwWidget *widget,
                                      DwRequisition *size)
{
   DwRequisition requisition;
   gint32 avail_width, avail_ascent, avail_descent;

   /* In future, boxes will probably be used for indented text. */
   avail_width =
      page->avail_width - Dw_style_box_diff_width(DW_WIDGET(page)->style) -
      widget->style->left_indent_first - widget->style->right_indent;
   avail_ascent =
      page->avail_ascent - Dw_style_box_diff_height(DW_WIDGET(page)->style);
   avail_descent = page->avail_descent;

   if (DW_WIDGET_USES_HINTS (widget)) {
      a_Dw_widget_set_width (widget, avail_width);
      a_Dw_widget_set_ascent (widget, avail_ascent);
      a_Dw_widget_set_descent (widget, avail_descent);
      a_Dw_widget_size_request (widget, size);
   } else {
      if (widget->style->width == DW_STYLE_UNDEF_LENGTH ||
          widget->style->height == DW_STYLE_UNDEF_LENGTH)
         a_Dw_widget_size_request (widget, &requisition);
   
      if (widget->style->width == DW_STYLE_UNDEF_LENGTH)
         size->width = requisition.width;
      else if (DW_STYLE_IS_LENGTH (widget->style->width))
         size->width =
            DW_STYLE_GET_LENGTH (widget->style->width, widget->style->font)
            + Dw_style_box_diff_width (widget->style);
      else
         size->width =
            DW_STYLE_GET_PERCENTAGE (widget->style->width) *  avail_width;
   
      if (widget->style->height == DW_STYLE_UNDEF_LENGTH) {
         size->ascent = requisition.ascent;
         size->descent = requisition.descent;
      } else if (DW_STYLE_IS_LENGTH (widget->style->width)) {
         size->ascent =
            DW_STYLE_GET_LENGTH (widget->style->height, widget->style->font)
            + Dw_style_box_diff_height (widget->style);
         size->descent = 0;
      } else {
         size->ascent =
            DW_STYLE_GET_PERCENTAGE (widget->style->height) * avail_ascent;
         size->descent = 
            DW_STYLE_GET_PERCENTAGE (widget->style->height) * avail_descent;
      }
   }
}


/*
 * Rewrap the page.
 * There are basically two times we'll want to do this:
 * either when the viewport is resized, or when the size changes on one
 * of the child widgets.
 */
static void Dw_page_rewrap(DwPage *page)
{
   DwPageLine *old_lines;
   DwWidget *widget;
   gint old_num_lines;
   gint line_index, word_index;
   gint nl;
   DwPageLine *old_line;
   DwPageWord *word;

   widget = DW_WIDGET (page);

   old_lines = page->lines;
   old_num_lines = page->num_lines;

   //g_print(">>>       Dw_page_rewrap (%d x %d x %d)\n",
   //page->avail_width, page->avail_ascent, page->avail_descent);

   /* Initialize the lines structure of the page. */
   page->num_lines = 0;
   page->num_lines_max = 1;
   page->lines = NULL;
   
   page->current_word = 0;
   
   page->real_width = 0;

   /* Iterate over the old lines, adding the words to the page structure. */
   for (line_index = 0; line_index < old_num_lines; line_index++) {
      old_line = &(old_lines[line_index]);
      /* Here, we're actually relying on the invariant that a soft line
       * is never followed by an empty hard line. */
      if (old_line->last_word == old_line->first_word && old_line->hard) {
         Dw_page_add_line(page);
      }
      for (word_index = old_line->first_word;
           word_index < old_line->last_word; word_index++) {
         word = &page->words[word_index];

         if (word->content_type == DW_PAGE_CONTENT_WIDGET)
            Dw_page_calc_widget_size (page, word->content.widget, &word->size);

         Dw_page_new_word(page, -1, -1, -1, NULL);
         
         if ( word->content_type == DW_PAGE_CONTENT_ANCHOR ) {
            Dw_widget_set_anchor(widget, word->content.anchor,
                                 page->lines[page->num_lines - 1].y_top);
         } 
      }
      
      if (old_line->hard) {
         a_Dw_page_linebreak(page);
         page->lines[page->num_lines - 1].y_space = old_line->y_space;
      }
   }
   g_free(old_lines);

   /* The last line hasn't been compared to real_width yet */
   nl = page->num_lines;
   if(nl > 0)
      page->real_width = MAX(page->real_width,
                             page->lines[nl - 1].size.width +
                             Dw_page_line_x_offset(page,
                                                   &(page->lines[nl - 1])));
}

/*
 * Paint a line
 * - x and y are toplevel dw coordinates (Question: what Dw? Changed. Test!)
 * - area is used always (ev. set it to event->area)
 * - event is only used when is_expose
 */
static void Dw_page_draw_line (DwPage *page,
                               DwPageLine *line,
                               DwRectangle *area,
                               GdkEventExpose *event)
{
   DwWidget *widget;
   DwPageWord *word;
   gint word_index;
   gint x_cursor, y_cursor;
   gint diff;
   gint uline_width;
   DwWidget *child;
   DwRectangle child_area;
   GdkWindow *window;

   /* Here's an idea on how to optimize this routine to minimize the number
    * of calls to gdk_draw_string:
    * 
    * Copy the text from the words into a buffer, adding a new word
    * only if: the attributes match, and the spacing is either zero or
    * equal to the width of ' '. In the latter case, copy a " " into
    * the buffer. Then draw the buffer. */

   widget = DW_WIDGET (page);
   window = DW_WIDGET_WINDOW (widget);
   x_cursor =
      Dw_widget_x_world_to_viewport (widget,
                                     widget->allocation.x +
                                     Dw_page_line_total_x_offset(page, line));
   y_cursor =
      Dw_widget_y_world_to_viewport (widget,
                                     widget->allocation.y
                                     + line->y_top + line->size.ascent);

   for (word_index = line->first_word; word_index < line->last_word; 
        word_index++) {
      word = &page->words[word_index];
      diff = 0;
      switch (word->content_type) {
      case DW_PAGE_CONTENT_TEXT:
         /* Adjust the text baseline if the word is <SUP>-ed or <SUB>-ed */
         if(word->style->SubSup == TEXT_SUB)
            diff = word->size.ascent / 2;
         else if(word->style->SubSup == TEXT_SUP)
            diff -= word->size.ascent / 3;

         gdk_draw_string(window,
                         word->style->font->font,
                         word->style->color->gc,
                         x_cursor,
                         y_cursor + diff,
                         word->content.text);
         if (word->style->uline >= 0) {
            uline_width = word->size.width;
            if (word_index + 1 < line->last_word &&
                word->style->uline == page->words[word_index + 1].style->uline)
               uline_width += word->x_space;
            gdk_draw_line(
               window,
               word->style->color->gc,
               x_cursor,
               y_cursor + 1 + diff,
               x_cursor + uline_width - 1,
               y_cursor + 1 + diff);
         }

         if (word->style->strike >= 0) {
            uline_width = word->size.width;
            if (word_index + 1 < line->last_word &&
                word->style->strike ==
                page->words[word_index + 1].style->strike)
               uline_width += word->x_space;
            gdk_draw_line(window,
                          word->style->color->gc,
                          x_cursor,
                          y_cursor - word->size.ascent / 2 + diff,
                          x_cursor + uline_width - 1,
                          y_cursor - word->size.ascent / 2 + diff);
         }
         break;

      case DW_PAGE_CONTENT_WIDGET:
         child = word->content.widget;
         if (Dw_widget_intersect (child, area, &child_area)) {
            //g_print ("Drawing widget %p\n", child);
            a_Dw_widget_draw (child, &child_area, event);
         }
         
         break;

      case DW_PAGE_CONTENT_ANCHOR:
         /* nothing - an anchor isn't seen */
         /* BUG: sometimes anchors have x_space; 
          * we subtract that just in case --EG */
         x_cursor -= word->size.width + word->x_space;
         break;
      }
      x_cursor += word->size.width + word->x_space;
   }
}

/*
 * Find the first line index that includes y, relative to top of widget.
 */
static gint Dw_page_find_line_index(DwPage *page, gint y)
{
   gint max_index = page->num_lines - 1;
   gint step, index, low = 0;

   step = (page->num_lines+1) >> 1;
   while ( step > 1 ) {
      index = low + step;
      if (index <= max_index && page->lines[index].y_top < y)
         low = index;
      step = (step + 1) >> 1;
   }

   if (low < max_index && page->lines[low+1].y_top < y)
      low++;

   /*
    * This new routine returns the line number between (y_top) and
    * (y_top + size.ascent + size.descent + y_space): the space
    * _below_ the line is considered part of the line.  Old routine
    * returned line number between (y_top - previous_line->y_space)
    * and (y_top + size.ascent + size.descent): the space _above_ the
    * line was considered part of the line.  This is important for
    * Dw_page_find_link() --EG
    */
   return low;
}

/*
 * Draw the actual lines, starting at (x, y) in toplevel Dw coords.
 * (former Dw_page_expose_lines)
 */
static void Dw_page_draw (DwWidget *widget,
                          DwRectangle *area,
                          GdkEventExpose *event)
{
   DwPage *page;
   gint line_index;
   DwPageLine *line;

   Dw_widget_draw_widget_box (widget, area);
  
   page = DW_PAGE (widget);
   line_index = Dw_page_find_line_index(page, area->y);

   for (; line_index < page->num_lines; line_index++) {
      line = &(page->lines[line_index]);
      if (line->y_top >= area->y + area->height)
         break;

      Dw_page_draw_line (page, line, area, event);
   }
}


/*
 * Find a link given a coordinate location relative to the window
 */
static gint Dw_page_find_link(DwPage *page, gint x, gint y)
{
   gint line_index, word_index;
   gint x_cursor, last_x_cursor;
   DwPageLine *line;
   DwPageWord *word;

   if ( (line_index = Dw_page_find_line_index(page, y)) >= page->num_lines )
      return -1;
   line = &page->lines[line_index];
   if (line->y_top + line->size.ascent + line->size.descent <= y)
      return -1;

   x_cursor = Dw_page_line_total_x_offset(page, line);
   for (word_index = line->first_word; word_index < line->last_word;
        word_index++) {
      word = &page->words[word_index];
      last_x_cursor = x_cursor;
      x_cursor += word->size.width + word->x_space;
      if (last_x_cursor <= x && x_cursor > x)
         return word->style->link;

   }
   return -1;
}


/*
 * Standard Dw function.
 */
static gint Dw_page_button_press  (DwWidget *widget,
                                   gint32 x,
                                   gint32 y,
                                   GdkEventButton *event)
{
   DwPage *page = DW_PAGE (widget);

#ifdef VERBOSE
   g_print("Dw_page_button_press: button (%d, %d) +%d\n",
           x, y, button->button);
#endif

   page->link_pressed = Dw_page_find_link(page, x, y);
   if (page->link_pressed >= 0) {
      gtk_signal_emit (GTK_OBJECT (widget), page_signals[LINK_PRESSED],
                       page->link_pressed, -1, -1, event);
      return TRUE;
   } else
      return FALSE;
}


/*
 * Standard Dw function.
 */
static gint Dw_page_button_release(DwWidget *widget,
                                   gint32 x,
                                   gint32 y,
                                   GdkEventButton *event)
{
   DwPage *page = DW_PAGE (widget);
   gint link_pressed, link_released;

#ifdef VERBOSE
   g_print("Dw_page_button_release: button (%d, %d) +%d\n",
           x, y, event->button);
#endif

   link_pressed = page->link_pressed;
   link_released = Dw_page_find_link(page, x, y);
   page->link_pressed = -1;

   if (link_released >= 0) {
      gtk_signal_emit (GTK_OBJECT (widget), page_signals[LINK_RELEASED],
                       link_released, -1, -1, event);

      if (link_pressed == link_released) {
         gtk_signal_emit (GTK_OBJECT (widget), page_signals[LINK_CLICKED],
                          link_released, -1, -1, event);
      }

      return TRUE;
   } else
      return FALSE;      
}


/*
 * Standard Dw function.
 */
static gint Dw_page_motion_notify (DwWidget *widget,
                                   gint32 x,
                                   gint32 y,
                                   GdkEventMotion *event)
{
   DwPage *page = DW_PAGE (widget);
   gint link, link_old;

   link_old = page->hover_link;
   link = Dw_page_find_link(page, x, y);
   page->hover_link = link;

   if (link != link_old) {
      gtk_signal_emit (GTK_OBJECT (widget), page_signals[LINK_ENTERED],
                       link, -1, -1);
      return TRUE;
   } else
      return (link != -1);  
}


/*
 * Standard Dw function.
 */
static gint Dw_page_leave_notify (DwWidget *widget,
                                  GdkEventMotion *event)
{
   DwPage *page = DW_PAGE (widget);
   
   if (page->hover_link != -1) {
      page->hover_link = -1;
      gtk_signal_emit (GTK_OBJECT (widget), page_signals[LINK_ENTERED],
                       -1, -1, -1);
   }
   
   return FALSE;
}


/*
 * Add a word to the page structure. Stashes the argument pointer in
 * the page data structure so that it will be deallocated on destroy.
 */
void a_Dw_page_add_text(DwPage *page, char *text, DwStyle *style)
{
   DwPageWord *word;
   gint width, ascent, descent;

   width = gdk_string_width(style->font->font, text);
   ascent = style->font->font->ascent;
   descent = style->font->font->descent;

   /* In case of a sub or super script we increase the word's height and
    * potentially the line's height. 
    */
   if(style->SubSup == TEXT_SUB)
      descent += (ascent / 2);
   else if(style->SubSup == TEXT_SUP)
      ascent += (ascent / 3);

   word = Dw_page_new_word(page, width, ascent, descent, style);
   word->content_type = DW_PAGE_CONTENT_TEXT;
   word->content.text = text;
   word->style = style;
   a_Dw_style_ref (style);
}

/*
 * Add a widget (word type) to the page. todo: comments.
 */
void a_Dw_page_add_widget (DwPage *page,
                           DwWidget *widget,
                           DwStyle *style)
{
   DwPageWord *word;
   DwRequisition size;

   a_Dw_widget_set_style (widget, style);

   Dw_page_calc_widget_size (page, widget, &size);
   word = Dw_page_new_word(page, size.width, size.ascent, size.descent, style);

   word->content_type = DW_PAGE_CONTENT_WIDGET;
   word->content.widget = widget;
   word->style = style;
   a_Dw_style_ref (style);

   Dw_widget_set_parent (widget, DW_WIDGET (page));
}


/*
 * Add an anchor to the page. name is copied, so no strdup is neccessary for
 * the caller.
 */
void a_Dw_page_add_anchor(DwPage *page, char *name, DwStyle *style)
{
   DwPageWord *word;
   
   word = Dw_page_new_word(page, 0, 0, 0, style);
   word->content_type = DW_PAGE_CONTENT_ANCHOR;
   word->content.anchor = g_strdup(name);
   word->style = style;
   a_Dw_style_ref (style);

   Dw_widget_set_anchor(DW_WIDGET(page), word->content.anchor,
                        page->lines[page->num_lines - 1].y_top);   
}


/*
 * ?
 */
void a_Dw_page_add_space(DwPage *page, DwStyle *style)
{
   gint nl, nw;
   gint space;

   nl = page->num_lines - 1;
   if (nl >= 0) {
      nw = page->num_words - 1;
      if (nw >= 0) {
         space = style->font->space_width;
         page->words[nw].x_space = space;
      }
   }
}

/*
 * Cause a line break
 */
void a_Dw_page_linebreak(DwPage *page)
{
   gint i;

   if ((i = page->num_lines) == 0)
      return;
   page->lines[i - 1].hard = TRUE;
}

/*
 * Cause a paragraph break
 */
void a_Dw_page_parbreak(DwPage *page, gint space)
{
   gint i;

   if ((i = page->num_lines) == 0)
      return;
   page->lines[i - 1].hard = TRUE;
   if (space == 0 || space > page->lines[i - 1].y_space)
      page->lines[i - 1].y_space = space;
}

/*
 * Call this routine before updating the state. By wrapping all state
 * changes between these two routines, the internal state of the widget
 * always matches the display, without requiring each state change
 * operation to calculate updates on a fine-grained basis.
 */
void a_Dw_page_update_begin(DwPage *page)
{
}


/*
 * Call this routine after updating the state.
 */
void a_Dw_page_update_end(DwPage *page)
{
   Dw_widget_queue_fast_resize (DW_WIDGET (page));
}


/*
 * Find the text in the page.
 * (Standar DwContainer function)    -- todo: move near forall
 */
static gint Dw_page_findtext(DwContainer *container, 
                             gpointer FP, gpointer KP,
                             gchar *NewKey)
{
   gint i;
   DwPageWord *word;
   FindData *F, *Fi, *Fj;
   DwPage *page = DW_PAGE(container);

   g_return_val_if_fail ((DW_WIDGET(page))->viewport != NULL, 0);
   if (!NewKey || !*NewKey)
      return 0;

   if ( !(F = *(FindData**)FP) )
      *(FindData **)FP = F = g_new0(FindData, 1);

   if ( !F->Key || strcmp(F->Key->KeyStr, NewKey) ||
        F->widget != DW_WIDGET (page) )
      F->State = F_NewKey;

   /* Let the FSM find the search string */
   while ( 1 ) {
      switch (F->State) {
      case F_NewKey:
         /* free FindData stack */
         for (Fi = F->next; Fi; Fi = Fj){
            Fj = Fi ? Fi->next : NULL;
            g_free(Fi);
         }
         /* initialize F */
         if (!KP && F->Key)
            a_Findtext_key_free(F->Key);
         F->Key = (KP) ? KP : a_Findtext_key_new(NewKey);
         F->widget = DW_WIDGET (page);
         F->WordNum = 0;
         F->next = NULL;
         F->State = F_Seek;
         break;
      case F_Seek:
         for (   ; F->WordNum < page->num_words; F->WordNum++) {
            word = &page->words[F->WordNum];
            if (word->content_type == DW_PAGE_CONTENT_TEXT &&
                a_Findtext_compare(word->content.text, F->Key)) {
               F->State = F_GetPos;
               break;
            } else if (word->content_type == DW_PAGE_CONTENT_WIDGET &&
                       DW_IS_CONTAINER (word->content.widget)) {
               if ( a_Dw_container_findtext(
                       DW_CONTAINER(word->content.widget),
                       (gpointer)&F->next, (gpointer)F->Key, NewKey) ) {
                  F->State = F_Seek;
                  return 1;
               }
            }
         }
         if (F->WordNum == page->num_words)
            F->State = F_End;
         break;
      case F_GetPos:
         if (F->Key->y_pos < 0) {
            for (i = 0; page->lines[i].last_word <= F->WordNum; ++i);
            F->Key->y_pos = page->lines[i].y_top;
         }
         F->WordNum++;
         F->State = (g_slist_length(F->Key->WordList) == F->Key->Matches) ?
                    F_Found : F_Seek;
         break;
      case F_Found:
         a_Dw_widget_scroll_to(DW_WIDGET (page), F->Key->y_pos);
         //g_print(">>>[Wn %d]\n", F->WordNum);
         F->State = F_Seek;
         return 1;
      case F_End:
         /* free memory */
         if (!KP && F->Key)
            a_Findtext_key_free(F->Key);
         g_free(F);
         *(FindData **)FP = NULL;
         return 0;
      }
   }

   /* compiler happiness */
   return 0;
}
